# -*- coding: utf-8 -*-
# Project : flacon
# File : trackset.py
# (c) Alexander Sokolov sokoloff.a@gmail.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


from settings import *
from flaconglobal import *
from track import *

from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import *

from errors import *

class TrackSet(QObject):
   #****************************************************************
   #
   #****************************************************************
    def __init__(self, uri = ''):
        QObject.__init__(self)
        self._tracks = []
        self._startTrackNum = 1
        self._discId = QString()
        self._textCodec = QTextCodec.codecForName(settings.value("Tags/Codepage").toString().toLatin1())
        self._uri = QString(uri)
        self._disc = None
        self._fileTag = QString()
        self._title = QString()
        self._titleEncoded = True


   #****************************************************************
   #
   #****************************************************************
    def uri(self):
        return self._uri


   #****************************************************************
   #
   #****************************************************************
    def setUri(self, value):
        self._uri = QString(value)


   #****************************************************************
   #
   #****************************************************************
    def title(self):
        if self._titleEncoded:
            return self._title
        else:
            return self.decode(self._title)


   #****************************************************************
   #
   #****************************************************************
    def setTitle(self, value, encoded = True):
        self._title = QString(value)
        self._titleEncoded = encoded


   #****************************************************************
   #
   #****************************************************************
    def textCodec(self):
        return self._textCodec


   #****************************************************************
   #
   #****************************************************************
    def decode(self, value):
        if not isinstance(value, QString):
            return value

        if (self._textCodec):
            return self._textCodec.toUnicode(value.toAscii())
        else:
            return value


   #****************************************************************
   #
   #****************************************************************
    def isActive(self):
        return (self._disc and self._disc.tracks() == self)


   #****************************************************************
   #
   #****************************************************************
    def activate(self):
        if self._disc:
            self._disc.selectTrackSet(self)


   #****************************************************************
   #
   #****************************************************************
    def trackDataChanged(self, track):
        if self._disc and self._disc.tracks() == self:
            self._disc.trackDataChanged(self.indexOf(track))


   #****************************************************************
   #
   #****************************************************************
    def __getitem__(self, index):
        return self._tracks[index]


   #****************************************************************
   #
   #****************************************************************
    def newTrack(self):
        track = Track(self)
        self._tracks.append(track)
        return track


   #****************************************************************
   #
   #****************************************************************
    def startTrackNum(self):
        return self._startTrackNum


   #****************************************************************
   #
   #****************************************************************
    def setStartTrackNum(self, value):
        self._startTrackNum = value
        for track in self._tracks:
            track.setStartTrackNum(value)


   #****************************************************************
   #
   #****************************************************************
    def count(self):
        return len(self._tracks)


   #****************************************************************
   #
   #****************************************************************
    def indexOf(self, track):
        try:
            return self._tracks.index(track)
        except:
            return -1


   #****************************************************************
   #
   #****************************************************************
    def codePage(self):
        return QString(self._textCodec.name())


   #****************************************************************
   #
   #****************************************************************
    def setCodePage(self, value):
        self._textCodec = QTextCodec.codecForName(QString(value).toLatin1())

        for i in xrange(0, self.count()):
            self.trackDataChanged(i)


   #****************************************************************
   #
   #****************************************************************
    def discId(self):
        return self._discId


   #****************************************************************
   #
   #****************************************************************
    def setDiscId(self, value):
        self._discId = value
        for i in xrange(0, self.count()):
            self.trackDataChanged(i)


   #****************************************************************
   #
   #****************************************************************
    def disc(self):
        return self._disc


   #****************************************************************
   #
   #****************************************************************
    def setDisc(self, value):
        self._disc = value


   #****************************************************************
   #
   #****************************************************************
    def fileTag(self):
        return self._fileTag


class CueTrackSet(TrackSet):
   #****************************************************************
   #
   #****************************************************************
    def __init__(self, fileName):
        TrackSet.__init__(self)
        self._uri = QFileInfo(fileName).canonicalFilePath()

        FileError.checkFile(fileName)
        f = QFile(fileName)

        self._title = QFileInfo(f).fileName()

        if f.open(QIODevice.ReadOnly | QIODevice.Text):
            self._cueData = QStringList()

            while not f.atEnd():
                self._cueData.append(QString(f.readLine()))

            #****************************************************
            # Delete BOM
            # see http://en.wikipedia.org/wiki/Byte_Order_Mark
            #****************************************************
            if (self._cueData[0].startsWith("\xEF\xBB\xBF")):   # UTF-8
                self.setCodePage('UTF-8')
                self._cueData[0] = self._cueData[0].remove(0, 3)

            if (self._cueData[0].startsWith("\xFE\xFF")):       # UTF-16 (BE)
                self.setCodePage('UTF-16BE')
                self._cueData[0] = self._cueData[0].remove(0, 2)

            if (self._cueData[0].startsWith("\xFF\xFE")):       # UTF-16 (LE)
                self.setCodePage('UTF-16LE')
                self._cueData[0] = self._cueData[0].remove(0, 2)

            self.parseCue(self._cueData)
            f.close()

            if not self.count():
                raise FileErrorBadCue(fileName)
        else:
            raise


   #****************************************************************
   #
   #****************************************************************
    def cueData(self):
        return self._cueData


   #****************************************************************
   #
   #****************************************************************
    def parseCue(self, lines):
        track  = QString()
        album  = QString()
        performer = QString()
        genre  = QString()
        discId = QString()
        year   = 0

        for i in xrange(0, lines.count()):
            line = lines[i].trimmed()

            if (line.isEmpty()): next
            line = line.trimmed()

            tag = line.section(' ', 0, 0).toUpper()
            value = line.section(' ', 1)
            if (value.length()>2 and (value[0] == "'" or value[0] == '"')):
                value = value.mid(1, value.length()-2)

            #==============================================
            if (tag == "TRACK"):
                num = value.section(' ', 0, 0)
                track = self.newTrack()
                track.setTrackNum(int(num))
                track.setArtist(performer, False)
                track.setAlbum(album, False)
                track.setGenre(genre, False)
                track.setYear(year)
                self.setDiscId(discId)

            #==============================================
            elif (tag == "TITLE"):
                if (track):
                    track.setTitle(value, False)
                else:
                    album = QString(value)

            #==============================================
            elif (tag == "PERFORMER"):
                if (track):
                    track.setArtist(value, False)
                else:
                    performer = QString(value)

            #==============================================
            elif (tag == "DATE"):           # For incorrect CUE files
                if (track):
                    track.setYear(value, False)
                else:
                    year = value

            #==============================================
            elif (tag == "FILE"):
                if not track:
                    self._fileTag = QString(value[:-5])
                else:
                    raise FileErrorMultiFileTag(self._uri)

            #==============================================
            elif (tag == "INDEX"):
                if track:
                    indexNum = value.section(' ', 0, 0)
                    time = value.section(' ', 1)

            #==============================================
            elif (tag == "REM"):
                subTag = value.section(' ', 0, 0).toUpper()
                value = value.section(' ', 1)
                if (value.length()>2 and (value[0] == "'" or value[0] == '"')):
                    value = value.mid(1, value.length()-2)

                if (not track):
                    if (subTag == 'DATE'):   year   = value
                    if (subTag == 'GENRE'):  genre  = value
                    if (subTag == 'DISCID'): discId = value


